@@ -26,29 +26,26 @@
 # Robin.antiX 04/20/2021 bugfix: check for remaining devices, did still not reliably detect leftover mountpoints.
 # Robin.antiX 04/20/2021 bugfix: deriving amount of time needed for sync from system info instead of fixed value; moreover before the delay window was not closing after sync due to “$!” mismatch of yad pids.
 # Robin.antiX 05/04/2021 new feature: "centred" optional for users preferring opening windows near mouse pointer. Additional command line option: -c or --centred.
-# Robin.antiX 05/06/2021 bugfix: replaced "pumount" command by "umount" Pumount has proven unreliable in dealing with partitions mounted on multiple mount points. Other mount points are unmounted than requested, causing the unmounting process for the next mountpoint of the partition to fail. Drawback: User will need to enter sudo password now.
+# Robin.antiX 05/06/2021 bugfix: replaced „pmount” by „umount” command for actual unmounting devices for several issues (pmount mixing up mountpoints for identical partitions and unfounded "Error: could not determine real path of the device" messages.)
+# Robin.antiX 06/08/2021 bugfix: mechanical (rotational) USB hdds did not spun down after unmounting, so it was not safe to unplug them wheras script tells user it was. Therefore it was neccessary to change default behaviour from “unsafe” to “safe” so user has to deactivate support for spinning devices actively by passing -u or -i option.
 
 # fixme: devices mounted to mountpoints containing blank characters still don't unmount properly.
-# fixme: mechanical USB hdds not spun down after unmounting, so it is not safe to unplug them when script tells user it was.
 
 ### Please Note: text strings for debug output dont't need to get marked for translation.
 
 
 # preliminaries
-title=$"Unplug USB device"
-version="0.82c"
-txt_version=$"Version "$version
-
 TEXTDOMAINDIR=/usr/share/locale
-TEXTDOMAIN=unplugdrive
+TEXTDOMAIN=unplugdrive.sh
 
-decorations="--undecorated"
-centred=""
-flag_safemode="FALSE"
+title=$"Unplug USB device"
+version="0.83"
+txt_version=$"Version "$version
 
 icon_taskbar="/usr/share/icons/numix-square-antix/48x48/devices/usbpendrive_unmount.png"
 declare -a mountpointlist
 #TEMPFILE=/tmp/unpluglist    # no longer needed
+#debuglog="$HOME/unplugdrive-debug-$(date +%d%m%Y%H%M%S).log"	# function to write a logfile not implemented yet
 removablelist=""
 removablenow=""
 summarylist=""
@@ -56,34 +53,46 @@ drivenow=""
 driveprevious=""
 drivelist=""
 deviceslist=()
+spindownlist=""
+spinningdownlist=""
+txt_rotationalerror=""
 umountfailurelist=()
 flag_debugmode=0
 flag_dryrun=0
+flag_spindown=0
 flag_umounterror=0
+flag_ignore_rotational=0
+decorations="--undecorated"
+centred=""
+flag_safemode="TRUE"
 CHR_0x0A="$(printf '\nx')"; CHR_0x0A=${CHR_0x0A%x}		# create special seperator
+txt_on=$"on" 											# refers to message “<partition> on <mountpoint>” presented in dialog to make it translatable.
 true_user=$SUDO_USER
 
+
 # check for commandline options
 txt_cline="$*"
 while [[ $# -gt 0 ]]
 do
   opt_cline="$1"
-
   case $opt_cline in
     -h|--help)	      echo ""
                       echo $"  Unplugdrive" "($txt_version)"
                       echo ""
                       echo -e $"  GUI tool for safely unplugging removable USB devices."
                       echo ""
-                      echo $"  Usage: sudo unplugdrive.sh [options]"
+                      echo $"  Usage: unplugdrive.sh [options]"
                       echo ""
                       echo $"  Options:"
                       echo -e $"\t-h or --help\t\tdisplays this help text"
-                      echo -e $"\t-s or --safe\t\truns script in safe mode,\n\t            \t\tdisplaying some extra dialogs"
+                      echo -e $"\t-u or --unsafe\t\truns script in unsafe mode,\n\t            \t\tomiting some checks and dialogs"
+                      echo -e $"\t-i or --ignore\t\tignores whether devices are\n\t            \t\treported as rotational even in\n\t            \t\tdefault mode. No spindown."
                       echo -e $"\t-d or --decorated\tuses window decorations"
                       echo -e $"\t-c or --centred\t\topen dialogs centred"
                       echo -e $"\t-g or --debug\t\tenable debug output (terminal)"
                       echo -e $"\t-p or --pretend\t\tdry run, don't actually un-\n\t            \t\tmount drives (for debugging)"
+                      echo -e ""
+                      echo -e $"  NEVER use the options -u and -i on rotational devices!"
                       echo ""
                       echo -e $"  Questions, Suggestions and Bugreporting please to:"
                       echo -e "\t<forum.antiXlinux.com>"
@@ -110,7 +119,9 @@ do
                       shift;;
     -c|--centred)     centred="--center"
                       shift;;                      
-    -s|--safe)        flag_safemode="TRUE"
+    -u|--unsafe)      flag_safemode="FLASE"
+                      shift;;
+    -i|--ignore)      flag_ignore_rotational=1
                       shift;;
     -g|--debug)       flag_debugmode=1
                       shift;;
@@ -121,7 +132,8 @@ do
                       shift;;
   esac
 done
-[ $flag_debugmode == 1 ] && echo -e "Version:\t$version\nDebug mode: on""\n$txt_cline\nsafemode:\t$flag_safemode\ndecorations:\t$decorations\ncentred:\t$centred\ndryrun (pretend):\t$flag_dryrun\nTrue user:\t$true_user\nexecuted by:\t`whoami`"
+
+[ $flag_debugmode == 1 ] && echo -e "Version:\t$version\nDebug mode: on""\n$txt_cline\nsafemode:\t$flag_safemode\ndecorations:\t$decorations\ncentred:\t$centred\ndryrun (pretend):\t$flag_dryrun\nignore rotationals\t$flag_ignore_rotational\nTrue user:\t$true_user\nexecuted by:\t`whoami`"
 
 # check whether unplugdrive is run with sudo by normal user.
 if [ `whoami` != "root" ]; then
@@ -158,7 +170,7 @@ return 0
 }
 
 # 1.) Preparations for user dialog:
-#------------------------------
+# ---------------------------------
 
 # Collect details of each removable device that has at least one mounted partition.
 # This is the base list for processing
@@ -179,7 +191,7 @@ done
 # This list is informational only, for display of mounted partitions in yad dialog!
 [ $flag_debugmode == 1 ] && echo "1.2.) tailored list for display purposes (\$removablenow):"
 for item in $detectedlist;do
-		removablenow=$(echo $item | cut -c 6- | sed 's:/.*/:/:' | sed 's/[/]//g')
+		removablenow="$(echo $item | cut -c 6- | sed 's:/.*/:/:' | sed 's/[/]//g' | sed 's/ on /\t'$txt_on'\t/g')"
 		[ $flag_debugmode == 1 ] && echo -e "removablenow item:\t$removablenow"
    removablelist="$removablelist$removablenow\n"
 done
@@ -258,9 +270,10 @@ if [ $flag_debugmode == 1 ]; then
 		done
 	echo -e "Number of items to get unmounted:\t${#mountpointlist[@]}"
 	echo -e "summary list is (\$summarylist):\n$summarylist"
-	echo -e "Unique devices to check for after unmount (\$deviceslist):\t$deviceslist"
+	echo -e "Unique devices to check for after unmount (\$deviceslist):\t${deviceslist[*]}"
 fi
 
+
 # Obtain confirmation to proceed with unmounting (only in safemode)
 if [ "$flag_safemode" = "TRUE" ]; then
 	sudo -u $true_user -- yad --title="$title" --fixed $centred --mouse $decorations \
@@ -295,12 +308,18 @@ sudo -u $true_user -- sync
 
 # wait and close delay window only after sync is done (only in safe mode)
 if [ "$flag_safemode" = "TRUE" ]; then
+	i=0
 	while :;
 	do
 		if [ "$(cat /proc/meminfo | grep 'Dirty:' | tr -s ' ' | cut -d' ' -f 2)" == "0" ] && [ "$(cat /proc/meminfo | grep 'Writeback:' | tr -s ' ' | cut -d' ' -f 2)" == "0" ]; then
 			if [ $flag_debugmode == 1 ]; then echo "sync done"; fi; break
-			fi
-	sleep 1				# wait until sync is done.
+		fi
+		sleep .2				# wait until sync is done before next check.
+		i=$[$i+1]
+		if [ $i -ge 5 ]; then
+			i=0
+			sudo -u $true_user -- sync		#feed some additional sync commands in case first check was missed on systems under heavy load.
+		fi
 	done
 	[ $flag_debugmode == 1 ] && echo -e "pids of existing yad processes (\$yad_pid_before):\n$yad_pid_before\npid of yad info window (\$yad_pid):\n$yad_pid"
 	kill $yad_pid
@@ -311,27 +330,28 @@ fi
 # -------------------------------
 
 # unmount each item in $mountpointlist
-[ $flag_debugmode == 1 ] && echo -e "Number of items in \$mountpointlist():  ${#mountpointlist[@]}"
+[ $flag_debugmode == 1 ] && echo -e "4.) Unmounting\nNumber of items in \$mountpointlist():  ${#mountpointlist[@]}"
 [ $flag_dryrun == 1 ] && echo -e "\033[1;33mDry run. Nothing was actually unmounted.\nOmit option -p for real operation.\033[0m"
 i=0
 while [ $i -lt ${#mountpointlist[@]} ]
 do
 	[ $flag_debugmode == 1 ] && echo "unmounting item $i is ${mountpointlist[i]}"
-	[ $flag_dryrun != 1 ] && gksudo -- umount "$(echo ${mountpointlist[$i]})"						# here selected devices get unmounted actually. (it needs 'echo' for some reason)
+	[ $flag_dryrun != 1 ] && umount "$(echo ${mountpointlist[$i]})"						# here selected devices get unmounted actually. (it needs 'echo' for some reason)
 	i=$[$i+1]
 done
 
 
 
-# 5.) Test for remaining mounted partitions and final user dialog on exit
+# 5.) Test for remaining mounted partitions, check for rotational devices 
 # -----------------------------------------------------------------------
 
 # check $deviceslist against all currtently mounted devices using findmnt command to get reliable data. Refering to “man mount”: >>For more robust and customizable output use findmnt(8), especially in your scripts.<<
+[ $flag_debugmode == 1 ] && echo -e "5.) Check for remaining unmounted partitions\nand check whether unmounted devices need spindown."
 u=0
 while [ $u -lt ${#deviceslist[@]} ]
 	do
 		checkitem="$(sudo -u $true_user -- findmnt --list -o SOURCE,TARGET | grep ${deviceslist[u]})"
-		[ $flag_debugmode == 1 ] &&  echo -e "checking for item $u:\n${deviceslist[u]}"
+		[ $flag_debugmode == 1 ] &&  echo -e "checking for item $u in \$deviceslist[n]:\n${deviceslist[u]}"
 		if [ -n "$checkitem" ]; then
 			umountfailurelist+=("${deviceslist[u]}")
 			echo -e $"\033[1;31mWARNING: DEVICE ${deviceslist[u]} WAS NOT PROPERLY UNMOUNTED!\033[0m\n\033[4;37mPlease check before unplugging.\033[0m"
@@ -343,11 +363,84 @@ while [ $u -lt ${#deviceslist[@]} ]
 					let $((v++))
 				done
 		else
-		[ $flag_debugmode == 1 ] &&  echo -e "Check OK"
+			[ $flag_debugmode == 1 ] &&  echo -e "Check OK"
+			# check which devices are reported as rotational and need to spin down.
+			if [ $(cat /sys/block/$(echo ${deviceslist[u]} | cut -d '/' -f 3)/queue/rotational) == 1 ]; then
+				[ $flag_debugmode == 1 ] &&  echo -e "Device $(echo ${deviceslist[u]} | cut -d '/' -f 3) is reported as “rotational”. Needs spindown."
+				spindownlist="$spindownlist${deviceslist[u]} "  # hint: since devcices in /dev never contain blanks it is safe here to use it as separator.
+				flag_spindown=1
+			fi
 		fi
 		let $((u++))
     done
 
+
+# 6.) Spin down rotational devices and wait for them reporting to be spun down.
+# -----------------------------------------------------------------------------
+
+# spin down device if it is marked as rotating. Otherwise it is still NOT safe to unplug. (in safemode only)
+if [ $flag_ignore_rotational == 1 ]; then flag_spindown=0; [ $flag_debugmode == 1 ] && echo -e "Ignoring rotational devices in need for spindown per user request"; fi
+if [ "$flag_safemode" = "TRUE" ] && [ $flag_spindown == 1 ]; then
+		echo -e $"Please wait until device(s) spun down..."
+		# get yad pid of specific yad window ($! will fail on yad windows)
+		yad_pid_before="$(pidof -x yad) "			# blank suffix intentionally
+		sudo -u $true_user -- yad --title="$title" --fixed $centred --mouse $decorations \
+			--text=$"Rotational devices\nspin down. <b>Please wait...</b>" \
+			--window-icon="$icon_taskbar" \
+			--no-buttons &
+		sleep 1										# wait for second yad pid
+		yad_pid="$(pidof -x yad)"
+		if [ "$yad_pid_before" != " " ]; then		# if no items sed will fail
+			while read -d' ' item
+			do
+				yad_pid="$(echo "$yad_pid" | sed "s/$item//g")"
+			done <<<"$yad_pid_before"
+		fi
+		[ $flag_debugmode == 1 ] && echo -e "6.1) Spinning down rotationals\nItems in spindownlist: $(echo "$spindownlist" | wc -w)"
+		while read -d ' ' item; do
+			hdparm -Y "$item"					# we can safely spin down rotational devices even if still mounted (can only happen in dryrun mode), they will power up again when accessed.
+			response=$?
+			if [ $response == 0 ];then
+				spinningdown=1
+				spinningdownlist="$spinningdownlist$item "
+			else
+				[ $flag_debugmode == 1 ] && echo -e "Item does not respond correctly to spindown command: $item\nItem removed from list."
+				txt_rotationalerror="$txt_rotationalerror$item "
+			fi
+		done <<<"$spindownlist"
+		[ "$(echo "$spinningdownlist" | wc -w)" == 0 ] && flag_spindown=0
+		[ $flag_debugmode == 1 ] && echo -e "6.2) Waiting for devices to report spun down\nItems in spinningdownlist: $(echo "$spinningdownlist" | wc -w)\nspinningdownlist: $spinningdownlist\nflag_spindown: $flag_spindown"
+# Following lines deactivated due to incompatibility to some drives, which don't respond to status query once received spindown command until they get accessed again and are spun up.
+# So we'll have to rely on exit status of first spindown command only for now.
+#		# wait and check until drives are reported as spun down (only if device responds correctly)
+#		while [ $flag_spindown == 1 ]; do
+#			flag_spindown=0
+#			while read -d ' ' item; do
+#				[ $flag_debugmode == 1 ] && echo --- $item	---
+#				[ -e "$(hdparm -C $item | grep -o standby)" ] && flag_spindown=1
+#				[ $flag_debugmode == 1 ] && echo -e "item: $item \t $flag_spindown"
+#			done <<<"$spinningdownlist"
+#			sleep 1  # give drives some time to report
+#			#[ -e "$(ps | grep '$(echo $yad_pid | cut -d' ' -f 1) \| $(echo $yad_pid | cut -d' ' -f 2)')" ] && flag_spindown=0 # go on if a device doesn't report correctly and user is sure all devices are spun down.
+#		done
+		[ $flag_debugmode == 1 ] && echo -e "Spindown for all requested devices done.\npids of existing yad processes (\$yad_pid_before):\n$yad_pid_before\npid of yad info window (\$yad_pid):\n$yad_pid"
+		kill $yad_pid
+fi		
+
+
+# 7.) Final user dialog on exit
+# -----------------------------
+
+if [ "$txt_rotationalerror" != "" ]; then
+	successtext=$"<b>Unmounted:</b>\n$summarylist\nBut despite the following devices were reported to be rotational\nthey did not respond to spindown command:\n\t$txt_rotationalerror\nPlease check whether they are spun down before unplugging.\n<b>After rotational drives are spun down it is safe to unplug.</b>"
+	yad_timeout=""
+	yad_button="--button="OK""
+else
+	successtext=$"<b>Unmounted:</b>\n$summarylist\n<b>It is safe to unplug the drive(s)</b>"
+	yad_timeout="--timeout=5"
+	yad_button="--no-buttons"
+fi
+
 # User information on exit:
 if [ $flag_umounterror == 1 ];then
 	# Display a message if unmount failed
@@ -360,10 +453,10 @@ if [ $flag_umounterror == 1 ];then
 	cleanup && exit 1
 else
    # Display a message if unmount successful   
-	sudo -u $true_user -- yad --title="$title" --fixed $centred --timeout=5 --mouse $decorations \
-        --text=$"<b>Unmounted:</b>\n$summarylist\n<b>It is safe to unplug the drive(s)</b>" \
+	sudo -u $true_user -- yad --title="$title" --fixed $centred $yad_timeout --mouse $decorations \
+        --text="$successtext" \
         --window-icon="$icon_taskbar" \
-        --no-buttons
+        $yad_button
         # no need for checking button selection anymore
 	cleanup && exit 0
 fi
